package com.carrotsearch.hppcrt.generator;

import org.junit.Test;

public class TemplateOptionsTest
{
    @Test
    public void testInlines() {

        final TemplateOptions testInstance = new TemplateOptions(Type.GENERIC, null);

        //nothing for now
    }

}
