package com.carrotsearch.hppcrt.predicates;

/**
 * A predicate that applies to <code>KType</code> objects.
 */
/*! ${TemplateOptions.generatedAnnotation} !*/
public interface KTypePredicate<KType>
{
    boolean apply(KType value);
}
