package com.carrotsearch.hppcrt.generator.parser;

public final class Channels
{
    private Channels() {
        //nothing
    }

    public final static int LINE_COMMENT = 1000;
    public final static int BLOCK_COMMENT = 1001;
    public final static int JAVADOC_COMMENT = 1002;
}
