package com.carrotsearch.hppcrt.procedures;

/**
 * A procedure that applies to <code>KType</code> objects.
 */
/*! ${TemplateOptions.generatedAnnotation} !*/
public interface KTypeProcedure<KType>
{
    void apply(KType value);
}
